#include "Signal.h"
#include "Symbol.h"
#include "Value.h"
#include "Expr.h"

namespace AScript {

//-----------------------------------------------------------------------------
// SignalType
//-----------------------------------------------------------------------------
const TCHAR *GetSignalTypeName(SignalType sigType)
{
	static const struct {
		SignalType sigType;
		const TCHAR *name;
	} tbl[] = {
		{ SIGTYPE_None,		_T("none"),		},
		{ SIGTYPE_Error,	_T("error"),	},
		{ SIGTYPE_Break,	_T("break"),	},
		{ SIGTYPE_Continue,	_T("continue"),	},
		{ SIGTYPE_Return,	_T("return"),	},
	};
	for (int i = 0; i < NUMBEROF(tbl); i++) {
		if (tbl[i].sigType == sigType) return tbl[i].name;
	}
	return _T("unknown");
}

//-----------------------------------------------------------------------------
// ErrorType
//-----------------------------------------------------------------------------
const TCHAR *GetErrorTypeName(ErrorType errType)
{
	static const struct {
		ErrorType errType;
		const TCHAR *name;
	} tbl[] = {
		{ ERR_None,					_T("None")				},
		{ ERR_SyntaxError,			_T("SyntaxError")		},
		{ ERR_ArithmeticError,		_T("ArithmeticError")	},
		{ ERR_TypeError,			_T("TypeError")			},
		{ ERR_ZeroDivisionError,	_T("ZeroDivisionError")	},
		{ ERR_ValueError,			_T("ValueError")		},
		{ ERR_SystemError,			_T("SystemError")		},
		{ ERR_IOError,				_T("IOError")			},
		{ ERR_IndexError,			_T("IndexError")		},
		{ ERR_KeyError,				_T("KeyError")			},
		{ ERR_ImportError,			_T("ImportError")		},
		{ ERR_AttributeError,		_T("AttributeError")	},
	};
	for (int i = 0; i < NUMBEROF(tbl); i++) {
		if (tbl[i].errType == errType) return tbl[i].name;
	}
	return _T("unknown");
}

//-----------------------------------------------------------------------------
// Signal
//-----------------------------------------------------------------------------
Signal::Signal() : _pMsg(new Message()), _stackLevel(0)
{
}

Signal::Signal(const Signal &sig) :
						_pMsg(sig._pMsg), _stackLevel(sig._stackLevel + 1)
{
	if (_stackLevel > MAX_STACK_LEVEL) {
		SetError(ERR_SystemError, _T("stack level exceeds maximum (%d)"), MAX_STACK_LEVEL);
	}
}

Signal::Message::Message() : sigType(SIGTYPE_None),
	errType(ERR_None), errTypeSaved(ERR_None), pExprCause(NULL), pValue(new Value())
{
}

String Signal::GetErrString() const
{
	String str;
	const Expr *pExprCause = GetExprCause();
	str += GetErrorName();
	if (pExprCause != NULL && pExprCause->GetLineNo() > 0) {
		TCHAR buff[32];
		::_stprintf(buff, _T(" at line.%d: "), pExprCause->GetLineNo());
		str += buff;
	} else {
		str += _T(": ");
	}
	str += _pMsg->str.c_str();
	return str;
}

void Signal::SetError(ErrorType errType, const TCHAR *format, ...)
{
	va_list list;
	va_start(list, format);
	SetErrorV(errType, format, list);
	va_end(list);
}

void Signal::SetErrorV(ErrorType errType,
					const TCHAR *format, va_list list, const TCHAR *textPre)
{
	TCHAR str[2048];
	::_tcscpy(str, textPre);
	::_vstprintf(str + ::_tcslen(textPre), format, list);
	_pMsg->sigType = SIGTYPE_Error;
	_pMsg->errType = errType;
	_pMsg->str = str;
	*_pMsg->pValue = Value::Null;
}

}
